@@ -5,6 +5,8 @@ extern crate terminal_size;
 
 extern crate tokio;
 
+extern crate once_cell;
+
 extern crate dnsclient;
 
 #[cfg(feature = "json")]
@@ -36,7 +38,9 @@ use tokio::net::UnixStream;
 use tokio::fs;
 use tokio::net::TcpStream;
 use tokio::sync::mpsc;
-use tokio::time::{self, sleep, Instant};
+use tokio::time::{self, sleep};
+
+use once_cell::sync::Lazy;
 
 use dnsclient::r#async::DNSClient;
 use dnsclient::UpstreamServer;
@@ -54,6 +58,25 @@ const CARGO_PKG_AUTHORS: &str = env!("CARGO_PKG_AUTHORS");
 const DEFAULT_TIMEOUT_SECONDS: &str = "60";
 const SLEEP_INTERVAL: Duration = Duration::from_millis(500);
 
+static DNS_CLIENT: Lazy<DNSClient> = Lazy::new(|| {
+    let dns_servers = vec![
+        UpstreamServer::new(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(1, 1, 1, 1)), 53)),
+        UpstreamServer::new(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(8, 8, 8, 8)), 53)),
+        UpstreamServer::new(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(4, 4, 4, 4)), 53)),
+    ];
+
+    #[cfg(unix)]
+    let client = match DNSClient::new_with_system_resolvers() {
+        Ok(client) => client,
+        Err(_) => DNSClient::new(dns_servers),
+    };
+
+    #[cfg(windows)]
+    let client = DNSClient::new(dns_servers);
+
+    client
+});
+
 #[cfg_attr(feature = "json", derive(Deserialize))]
 #[derive(Debug)]
 struct TcpTask {
@@ -106,22 +129,7 @@ async fn host_port_to_socket_addrs(
     match IpAddr::from_str(host) {
         Ok(ip) => Ok(vec![SocketAddr::new(ip, port)]),
         Err(_) => {
-            let dns_servers = vec![
-                UpstreamServer::new(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(1, 1, 1, 1)), 53)),
-                UpstreamServer::new(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(8, 8, 8, 8)), 53)),
-                UpstreamServer::new(SocketAddr::new(IpAddr::V4(Ipv4Addr::new(4, 4, 4, 4)), 53)),
-            ];
-
-            #[cfg(unix)]
-            let client = match DNSClient::new_with_system_resolvers() {
-                Ok(client) => client,
-                Err(_) => DNSClient::new(dns_servers),
-            };
-
-            #[cfg(windows)]
-            let client = DNSClient::new(dns_servers);
-
-            Ok(client
+            Ok(DNS_CLIENT
                 .query_addrs(host)
                 .await?
                 .into_iter()
@@ -131,49 +139,39 @@ async fn host_port_to_socket_addrs(
     }
 }
 
-async fn wait_tcp(tcp_task: &TcpTask, timeout: Duration) -> Result<(), String> {
-    if timeout.is_zero() {
+async fn wait_tcp_handler(tcp_task: &TcpTask) -> Result<(), String> {
+    'outer: loop {
         let addrs = host_port_to_socket_addrs(tcp_task.host.as_str(), tcp_task.port)
             .await
             .map_err(|err| format!("Cannot resolve the host: {:?} {}", tcp_task.host, err))?;
 
-        'outer: loop {
-            for addr in addrs.iter().cloned() {
-                if TcpStream::connect(addr).await.is_ok() {
-                    break 'outer;
-                }
+        for addr in addrs.iter().cloned() {
+            if TcpStream::connect(addr).await.is_ok() {
+                break 'outer;
             }
-
-            sleep(SLEEP_INTERVAL).await;
         }
+
+        sleep(SLEEP_INTERVAL).await;
+    }
+
+    Ok(())
+}
+
+async fn wait_tcp(tcp_task: &TcpTask, timeout: Duration) -> Result<(), String> {
+    if timeout.is_zero() {
+        wait_tcp_handler(tcp_task).await?;
     } else {
-        let addrs = time::timeout(
-            timeout,
-            host_port_to_socket_addrs(tcp_task.host.as_str(), tcp_task.port),
-        )
-        .await
-        .map_err(|_| format!("Cannot resolve to host: {:?} timeout", tcp_task.host))?
-        .map_err(|err| format!("Cannot resolve the host: {:?} {}", tcp_task.host, err))?;
+        time::timeout(timeout, wait_tcp_handler(tcp_task)).await.map_err(|_| {
+            format!("Cannot connect to server: {}:{} timeout", tcp_task.host, tcp_task.port)
+        })??;
+    }
 
-        let start = Instant::now();
+    Ok(())
+}
 
-        'outer_timeout: loop {
-            for addr in addrs.iter().cloned() {
-                match time::timeout(timeout, TcpStream::connect(addr))
-                    .await
-                    .map_err(|_| format!("Cannot connect to the address: {} timeout", addr))?
-                {
-                    Ok(_) => break 'outer_timeout,
-                    Err(err) => {
-                        if Instant::now() - start > timeout {
-                            return Err(format!("Cannot connect to the address: {} {}", addr, err));
-                        }
-                    }
-                }
-            }
-
-            sleep(SLEEP_INTERVAL).await;
-        }
+async fn wait_uds_handler(uds_task: &UdsTask) -> Result<(), String> {
+    while UnixStream::connect(uds_task.uds.as_str()).await.is_err() {
+        sleep(SLEEP_INTERVAL).await;
     }
 
     Ok(())
@@ -182,31 +180,14 @@ async fn wait_tcp(tcp_task: &TcpTask, timeout: Duration) -> Result<(), String> {
 #[cfg(unix)]
 async fn wait_uds(uds_task: &UdsTask, timeout: Duration) -> Result<(), String> {
     if timeout.is_zero() {
-        while UnixStream::connect(uds_task.uds.as_str()).await.is_err() {
-            sleep(SLEEP_INTERVAL).await;
-        }
+        wait_uds_handler(uds_task).await?
     } else {
-        let start = Instant::now();
-
-        while let Err(err) = time::timeout(timeout, UnixStream::connect(uds_task.uds.as_str()))
-            .await
-            .map_err(|_| {
-                format!(
-                    "Cannot connect to the socket: {:?} timeout",
-                    Path::new(uds_task.uds.as_str()).absolutize().unwrap().to_string_lossy()
-                )
-            })?
-        {
-            if Instant::now() - start > timeout {
-                return Err(format!(
-                    "Cannot connect to the socket: {:?} {}",
-                    Path::new(uds_task.uds.as_str()).absolutize().unwrap().to_string_lossy(),
-                    err
-                ));
-            } else {
-                sleep(SLEEP_INTERVAL).await;
-            }
-        }
+        time::timeout(timeout, wait_uds_handler(uds_task)).await.map_err(|_| {
+            format!(
+                "Cannot connect to the socket: {:?} timeout",
+                Path::new(uds_task.uds.as_str()).absolutize().unwrap().to_string_lossy()
+            )
+        })??;
     }
 
     Ok(())
@@ -352,9 +333,9 @@ async fn main() -> Result<(), Box<dyn Error>> {
     exec(command)
 }
 
-fn get_matches<'a>() -> ArgMatches<'a> {
+fn get_matches() -> ArgMatches {
     let app = App::new(APP_NAME)
-        .set_term_width(terminal_size().map(|(width, _)| width.0 as usize).unwrap_or(0))
+        .term_width(terminal_size().map(|(width, _)| width.0 as usize).unwrap_or(0))
         .version(CARGO_PKG_VERSION)
         .author(CARGO_PKG_AUTHORS)
         .about(concat!("Wait Service is a pure rust program to test and wait on the availability of multiple services\n\nEXAMPLES:\n", concat_line!(prefix "wait-service ",
@@ -363,65 +344,64 @@ fn get_matches<'a>() -> ArgMatches<'a> {
             "--uds /var/run/app.sock --json /path/to/json       -- npm start  # Wait for /var/run/app.sock and other services defined in the json file (max 60 seconds) and then run `npm start`",
         )));
 
-    let arg_timeout = Arg::with_name("TIMEOUT")
+    let arg_timeout = Arg::new("TIMEOUT")
         .display_order(0)
-        .required(true)
         .long("timeout")
-        .short("t")
+        .short('t')
         .takes_value(true)
         .default_value(DEFAULT_TIMEOUT_SECONDS)
         .help("Sets the timeout in seconds, zero for no timeout");
 
-    let arg_command = Arg::with_name("COMMAND")
+    let arg_command = Arg::new("COMMAND")
         .required(true)
         .help("Command to execute after service is available")
-        .multiple(true);
+        .multiple_values(true);
 
-    let arg_tcp = Arg::with_name("TCP")
+    let arg_tcp = Arg::new("TCP")
         .display_order(1)
         .long("tcp")
         .takes_value(true)
         .help("Test and wait on the availability of TCP services")
-        .multiple(true);
+        .multiple_values(true);
 
     let arg_tcp = if cfg!(unix) {
         if cfg!(feature = "json") {
-            arg_tcp.required_unless_one(&["UDS", "JSON"])
+            arg_tcp.required_unless_present_any(&["UDS", "JSON"])
         } else {
-            arg_tcp.required_unless_one(&["UDS"])
+            arg_tcp.required_unless_present_any(&["UDS"])
         }
     } else if cfg!(feature = "json") {
-        arg_tcp.required_unless_one(&["JSON"])
+        arg_tcp.required_unless_present_any(&["JSON"])
     } else {
         arg_tcp
     };
 
     #[cfg(unix)]
-    let arg_uds = Arg::with_name("UDS")
+    let arg_uds = Arg::new("UDS")
         .display_order(2)
         .long("uds")
         .takes_value(true)
         .visible_alias("unix")
-        .required_unless_one(if cfg!(feature = "json") {
+        .required_unless_present_any(if cfg!(feature = "json") {
             ["TCP", "JSON"].as_ref()
         } else {
             ["TCP"].as_ref()
         })
         .help("Test and wait on the availability of UDS services")
-        .multiple(true);
+        .multiple_values(true);
 
     #[cfg(feature = "json")]
-    let arg_json = Arg::with_name("JSON")
+    let arg_json = Arg::new("JSON")
         .display_order(3)
         .long("json")
         .takes_value(true)
-        .required_unless_one(if cfg!(unix) {
+        .required_unless_present_any(if cfg!(unix) {
             ["TCP", "UDS"].as_ref()
         } else {
             ["TCP"].as_ref()
         })
         .help("Test and wait on the availability of TCP or UDS services")
-        .multiple(true);
+        .multiple_values(true);
 
     let app = app.arg(arg_timeout).arg(arg_command).arg(arg_tcp);
 
